//     Programmer: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:       University of Padova - Department of Physics
//                 Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:      +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:      loreti@pd.infn.it
//     WWW:        http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// The following code declares and defines classes intended for
// reading from, and writing to, files identified by an Unix file
// descriptor.  The original fdoutbuf/fdinbuf and fdostream/fdistream
// classes were written by Frank B. Brokken; they are taken from
// chapter 20 of its book "C++ annotations" and modified by me (MLO).
// The book itself is available on the web at
// http://www.icce.rug.nl/documents/cplusplus/cplusplus.html .
//
// Copyright informations:
//
// (C) Copyright Frank B. Brokken, 2004
// All rights reserved. No part of this document may be changed or
// published in any form without prior consent of the author.
// -------------------------------------------------------------------
// $Id: fdstream.hh,v 1.1 2004/11/24 18:08:13 loreti Exp $
// -------------------------------------------------------------------

#ifndef FDSTREAM_HH_
#define FDSTREAM_HH_

#include <istream>
#include <ostream>
#include <streambuf>

extern "C" {
  #include <unistd.h>
}

// fdostream: an output stream that will be written on a file
// descriptor.  Uses fdoutbuf as internal buffer.

class fdoutbuf : public std::streambuf {
protected:
  // Data members:
  // - file descriptor;
  // - buffer size;
  // - pointer to the output buffer;
  // - default buffer size (static).

  int                 _fd;
  size_t              _bufsize;
  char *              _buffer;
  static const size_t defBufsize = 1024;

public:
  // Constructors: the default constructor merely initializes all the
  // internal members to invalid values; normally a fdoutbuf will be
  // constructed from a valid file descriptor and a buffer size (that
  // defaults to defBufsize); use 0 or 1 for an unbuffered stream.

  fdoutbuf() : _fd(-1), _bufsize(0), _buffer(0) {}
  fdoutbuf(int fd, size_t bufsize = defBufsize) { init(fd, bufsize); }

  virtual ~fdoutbuf() { if (_buffer) { sync(); delete [] _buffer; }}

private:
  virtual void init(
    int    fd,
    size_t s
  ) {
    // An internal function that properly initializes a fdoutbuf

    _fd      = fd;
    _bufsize = s ? s : 1;
    _buffer  = new char[_bufsize];
    setp(_buffer, _buffer + _bufsize);
  }

protected:
  virtual int_type sync() {
    // A member function to override the one in std::streambuf

    size_t s;

    if ((s = pptr() - pbase()) > 0) {
      if (write(_fd, _buffer, s) != static_cast<ssize_t>(s)) {
        return -1;
      }
      setp(_buffer, _buffer + _bufsize);
    }
    return 0;
  }

  virtual int_type overflow(
    int_type c
  ) {
    // A member function that writes a single character

    sync();
    if (c != traits_type::eof()) {
      *pptr() = static_cast<char>(c);
      pbump(1);
    }
    return c;
  }

  virtual std::streamsize xsputn(
    const char *    from,
    std::streamsize num
  ) {
    // A member function that stores num characters in the output buffer

    std::streamsize nwritten = 0;

    while (num) {
      std::streamsize avail = epptr() - pptr();
      if (!avail   &&   sync() != 0) break;
      if (avail > num) avail = num;
      std::memcpy(pptr(), from + nwritten, avail);
      pbump(avail);

      nwritten += avail;
      num      -= avail;
    }
    return nwritten;
  }
};

class fdostream : public std::ostream {
protected:
  fdoutbuf buf;

public:
  fdostream(int fd) : std::ostream(0), buf(fd) { rdbuf(&buf); }
};

// fdistream: a stream that reads on a file descriptor.  Uses fdinbuf
// as internal buffer.

class fdinbuf : public std::streambuf {
protected:
  // Data members:
  // - file descriptor;
  // - buffer size;
  // - size of the "unget" buffer;
  // - pointer to the input buffer;
  // - address, in the input buffer, for read operations;
  // - default values for the sizes of the "unget" buffer and of the
  //   whole input buffer (static).

  int                 _fd;
  size_t              _bufsize;
  size_t              _reserved;
  char *              _buffer;
  char *              _base;
  static const size_t defReserved = 4;
  static const size_t defBufsize  = 1024;

public:
  // Constructors: the default constructor merely initializes all the
  // internal variables to invalid values; normally a fdinbuf will be
  // constructed from a valid file descriptor and the sizes for the
  // whole input buffer and for the putback area (the default values
  // are above).

  fdinbuf() : _fd(-1), _bufsize(0), _reserved(0), _buffer(0), _base(0) {}
  fdinbuf(int fd, size_t bufsize = defBufsize,
          size_t reserved = defReserved) { init(fd, bufsize, reserved); }

  virtual ~fdinbuf() { delete [] _buffer; }

private:
  virtual void init(
    int    fd,
    size_t bs,
    size_t res
  ) {
    // A member function that properly initializes a fdinbuf

    if (res == 0)  res = 1;
    if (bs <= res) bs  = res + 1;

    _fd       = fd;
    _bufsize  = bs - res;
    _reserved = res;
    _buffer   = new char[bs];
    _base     = _buffer + _reserved;
    setg(_base, _base, _base);
  }

protected:
  virtual int_type underflow() {
    // Inserts a character into the buffer.  Is the read position
    // before the end of buffer?

    if (gptr() < egptr()) return traits_type::to_int_type(*gptr());

    // If not, processes the size of the putback area: uses the number
    // of characters read, but at most the size of putback area.

    size_t move = gptr() - eback();
    if (move > _reserved) move = _reserved;

    // Copies up to "move" characters previously read into the putback
    // area ("memmove" properly handles moves between overlapping
    // memory areas).

    std::memmove(_base - move, egptr() - move, move);

    // Reads at most _bufsize new characters; if read returns -1, that
    // means or error or EOF

    int num = read(_fd, _base, _bufsize);
    if (num <= 0) return traits_type::eof();

    // Resets the buffer pointers and returns the next character

    setg(_base - move, _base, _base + num);
    return traits_type::to_int_type(*gptr());
  }

  std::streamsize xsgetn(
    char *          dest,
    std::streamsize num
  ) {
    // Reads "num" characters, avoiding multiple calls to read a
    // single one

    std::streamsize nread = 0;

    while (num) {
      std::streamsize avail = in_avail();
      if (!avail   &&   underflow() == traits_type::eof()) break;
      if (avail > num) avail = num;
      std::memcpy(dest + nread, gptr(), avail);
      gbump(avail);

      nread += avail;
      num   -= avail;
    }
    return nread;
  }
};

class fdistream : public std::istream {
protected:
  fdinbuf buf;

public:
  fdistream(int _fd) : std::istream(0), buf(_fd) { rdbuf(&buf); }
};

#endif // FDSTREAM_HH_
